from .entity import Entity

class NamedEntity(Entity):
    """ A Named Entity representing system/SIMOS/NamedEntity"""

    def __init__(self) -> None:
        self.__name = ""

    @property
    def name(self) -> str:
        """"""
        return self.__name

    @name.setter
    def name(self, value: str):
        """Set name"""
        self.__name = str(value)



"""
{"name": "x", "type": "float", "description": "point, x coordinates"},
"""
from .attribute import Attribute
from .dimension import Dimension

class BlueprintAttribute(Attribute):
    """ A property"""

    def __init__(self,name:str ,attribute_type:str,description:str, contained: bool, *dimensions: Dimension) -> None:
        super().__init__(name,attribute_type,description,*dimensions)
        self.__contained = contained

    @property
    def is_primitive(self) -> bool:
        """Is this a primitive attribute"""
        return False

    @property
    def contained(self) -> bool:
        return self.__contained

    @contained.setter
    def contained(self, value: bool):
        """Set uid"""
        self.__contained = value



"""
{"name": "x", "type": "float", "description": "point, x coordinates"},
"""
from .dimension import Dimension

class Attribute:
    """ An attribute"""

    def __init__(self,name:str ,attribute_type:str,description:str, *dimensions: Dimension,
                 optional=True, default=None) -> None:
        self.name = name
        self.__type = attribute_type
        self.description = description
        self.dimensions = dimensions
        self.optional = optional
        self.default = default

    def has_dimensions(self):
        """Has dimensions"""
        return self.dimensions and len(self.dimensions)>0

    def is_boolean(self):
        """Is this a boolean primitive?"""
        return self.__type == 'boolean'

    def is_string(self):
        """Is this a String primitive?"""
        return self.__type == 'string' or self.__type == 'char'

    @property
    def contained(self) -> bool:
        """Is this attribute contained"""
        return True

    @property
    def is_primitive(self) -> bool:
        """Is this a primitive attribute"""
        return True

    @property
    def type(self) -> str:
        return self.__type

    @property
    def is_enum(self) -> bool:
        """Is this an enum primitive?"""
        return False



"""
{"name": "x", "type": "float", "description": "point, x coordinates"},
"""
from .attribute import Attribute

class EnumAttribute(Attribute):
    """ An Enum attribute"""

    def __init__(self,name:str ,attribute_type:str,description:str) -> None:
        super().__init__(name,attribute_type,description)

    @property
    def contained(self) -> bool:
        return False

    @property
    def is_primitive(self) -> bool:
        """Is this a primitive attribute"""
        return False

    @property
    def is_enum(self) -> bool:
        return True



""" A basic Entity Istance"""

from __future__ import annotations
from enum import Enum

from typing import Iterator, Sequence

from .blueprint import Blueprint
from .dimension import Dimension
from .attribute import Attribute


class Entity():
    """ A basic Entity Istance"""

    def __init__(self, **kwargs):
        pass

    @property
    def blueprint(self) -> Blueprint:
        """Return blueprint that this entity represents"""
        raise Exception("Should have been overridden")

    def get_dimension(self, dim: Dimension) -> int:
        return getattr(self,dim.name,0)

    def is_set(self, prop: Attribute) -> bool:
        value=getattr(self,prop.name,None)
        if value is None:
            return False
        if prop.is_string():
            if isinstance(value,Enum):
                #FIXME
                return True
            return len(value) > 0
        if prop.has_dimensions():
            return len(value)>0
        return True

    def content(self) -> Iterator[Entity]:
        for p in self.blueprint.blueprint_attributes():
            if p.contained and self.is_set(p):
                value = getattr(self, p.name, None)
                if p.has_dimensions():
                    children: Sequence[Entity] = value
                    for child in children:
                        yield child
                else:
                    child: Entity = value
                    yield child

    def all_content(self) -> Iterator[Entity]:
        for p in self.blueprint.blueprint_attributes():
            if p.contained and self.is_set(p):
                value = getattr(self, p.name, None)
                if p.has_dimensions():
                    children: Sequence[Entity] = value
                    for child in children:
                        yield child
                        yield from child.all_content()
                else:
                    child: Entity = value
                    yield child
                    yield from child.all_content()

    def copy(self) -> Entity:
        from .dmt_reader import DMTReader
        from .dmt_writer import DMTWriter
        entity_dict = DMTWriter().to_dict(self)
        return DMTReader().from_dict(entity_dict)




"""
# 	{
# 	    "name": "nPoints",
# 	    "description": "number of points"
# 	}
"""

class Dimension:
    """ A dimension"""

    def __init__(self, name:str, description:str=None) -> None:
        self.name = name
        self.description = description

    def __add__(self, other):
        # Ignore for now..
        return self



""" Export entites as SIMA objects"""


from enum import Enum
import json
from typing import Dict, Sequence
import uuid
from .attribute import Attribute
from .blueprint_attribute import BlueprintAttribute
from .entity import Entity


class DMTWriter:
    """Convert to DMT dictionary"""

    def __init__(self):
        self.uuids = dict()

    def write(self, entity: Entity, filename, indent=0):
        """Write entity to file"""
        with open(filename, "w", encoding="utf-8") as file:
            res = self.to_dict(entity)
            json.dump(res, file, indent=indent)

    def to_dicts(self, entities: Sequence[Entity]) -> Sequence[Dict]:
        """Convert to DMT dictionaries"""

        # Make sure all referenced enitites has id's
        for entity in entities:
            self.__set_alls_ids(entity)

        return [self.__as_dict(entity) for entity in entities]

    def to_dict(self, entity: Entity) -> Dict:
        """Convert to DMT dictionary"""
        return self.to_dicts([entity])[0]

    def __as_dict(self, entity: Entity):
        """Convert to dictionary"""
        blueprint = entity.blueprint
        ret = {"type": blueprint.get_path()}
        for attribute in blueprint.attributes:
            if entity.is_set(attribute):
                value = self.__attribute_dict(entity, attribute)
                if value is not None:
                    ret[attribute.name] = value
        _id = self.uuids.get(entity, None)
        if _id:
            ret["_id"] = _id
        return ret

    def __attribute_dict(self, entity: Entity, attribute: Attribute):
        value = getattr(entity, attribute.name, None)
        if isinstance(attribute, BlueprintAttribute):
            if not attribute.contained:
                # This is a cross reference
                reference: Entity = value
                _id = self.uuids.get(reference, None)
                if not _id:
                    raise Exception("Id not set")
                return {"_id": _id}
            if attribute.has_dimensions():
                values = [self.__as_dict(lvalue) for lvalue in value]
                return values
            else:
                return self.__as_dict(value)
        else:
            if attribute.is_primitive:
                if self.__is_optional_default(attribute, value):
                    return None
                return value
            else:
                if attribute.is_enum:
                    enum: Enum = value
                    return enum.name
                return self.__as_dict(value)

    def __set_alls_ids(self, entity: Entity):
        for child in entity.all_content():
            for atribute in child.blueprint.blueprint_attributes():
                if not atribute.contained and child.is_set(atribute):
                    self.__set_id(child, atribute)

    def __set_id(self, entity: Entity, attribute: BlueprintAttribute):
        uuids = self.uuids
        value = getattr(entity, attribute.name, None)
        if attribute.has_dimensions():
            entities: Sequence[Entity] = value
            for entity in entities:
                if entity not in uuids:
                    uuids[entity] = str(uuid.uuid4())
        else:
            entity: Entity = value
            if entity not in uuids:
                uuids[entity] = str(uuid.uuid4())

    def __is_optional_default(self, attribute: Attribute, value: any):
        if attribute.optional and attribute.is_primitive:
            return value == attribute.default
        return False



""" Creates entities from Dicts """


from copyreg import constructor
import json
from importlib import import_module
from pyclbr import Function
from typing import Dict, Sequence

from .attribute import Attribute
from .enum_attribute import EnumAttribute
from .blueprint_attribute import BlueprintAttribute
from .entity import Entity


class DMTReader():
    """ Creates entities from Dicts """

    class Reference:
        """Holds a reference until it can be resolved"""

        def __init__(self,entity, prop: Attribute,uid: str):
            self.entity = entity
            self.prop = prop
            self.uid = uid

    def __init__(self):
        self.entities = dict()
        self.unresolved = list()

    def read(self, filename) -> Entity:
        """ Read entity from file """
        with open(filename, 'r',encoding="utf-8") as file:
            res=json.load(file)
            return self.from_dict(res)


    def from_dict(self,ent_dict: Dict) -> Entity:
        """ Create entity from Dict """
        entity=self.__from_dict(ent_dict)
        self.__resolve_all()
        return entity

    def from_dicts(self,ent_dicts: Sequence[Dict]) -> Sequence[Entity]:
        """ Create entity from Dict """
        entities = [self.__from_dict(e) for e in ent_dicts]
        self.__resolve_all()
        return entities

    def __resolve_all(self):
        for ref in self.unresolved:
            if not self.__resolve(ref):
                raise Exception(f"Unresolved reference: {ref}")


    def __from_dict(self,ent_dict: Dict) -> Entity:
        """ Read entities from Dict """
        entity_type: str=ent_dict["type"]
        constructor = self._resolve_type(entity_type)
        if not constructor:
            raise Exception(f"Unkown entity type {entity_type}")
        entity_instance: Entity = constructor()
        blueprint = entity_instance.blueprint
        for key, value in ent_dict.items():
            if key == "_id":
                uid = value
                self.entities[uid] = entity_instance
                continue
            if key == "type":
                continue
            attribute = blueprint.get_attribute(key)
            if not attribute:
                #FIXME
                continue
            if isinstance(attribute, BlueprintAttribute):
                self.__set_blueprint_value(entity_instance,attribute,value)
            elif attribute.is_enum:
                self.__set_enum_value(entity_instance,attribute,value)
            else:
                setattr(entity_instance,key, value)

        return entity_instance

    def _resolve_type(self, atype: str) -> Function:
        pkg: any = None
        parts = atype.split("/")
        if parts[0] == "":
            del parts[0]
        ename = parts.pop()
        package_path = ".".join(parts)
        try:
            pkg = import_module(package_path)
        except ModuleNotFoundError:
            raise Exception(f"Unable to load package {package_path}")
    
        constructor = pkg.__dict__.get(ename)
        return constructor

    def __set_blueprint_value(self,entity_instance: Entity, attribute: EnumAttribute,value):
        if attribute.contained:
            self.__set_value(entity_instance,attribute,value)
        else:
            self.__set_reference(entity_instance,attribute,value)

    def __set_reference(self,entity_instance: Entity,prop: Attribute,value):
        if isinstance(value, Dict):
            uid=value["_id"]
            ref = self.Reference(entity_instance,prop,uid)
            if not self.__resolve(ref):
                self.unresolved.append(ref)
        return

    def __set_value(self,entity_instance: Entity, attribute: BlueprintAttribute,value):
        if isinstance(value, Sequence):
            children = [self.__from_dict(v) for v in value]
            setattr(entity_instance,attribute.name, children)
        elif isinstance(value, Dict):
            child=self.__from_dict(value)
            if child:
                setattr(entity_instance,attribute.name, child)
        else:
            setattr(entity_instance,attribute.name, value)

    def __resolve(self, ref: Reference):
        value = self.entities.get(ref.uid)
        if value:
            self.__set_value(ref.entity,ref.prop,value)
            return True
        return False

    def __set_enum_value(self,entity: Entity, attribute: EnumAttribute,value: str):
        """ Convert from string to Enum"""
        
        constructor = self._resolve_type(attribute.type)
        if not constructor:
            raise Exception(f"Unkown Enum type {attribute.type}")
        evalue = constructor[value]
        setattr(entity,attribute.name, evalue)



import json
from dmt.attribute import Attribute
from dmt.blueprint import Blueprint
from dmt.dmt_writer import DMTWriter
from tests.test_entity import TestEntity


def test_attribute_creation():

    attribute = Attribute("myint", "integer", "description")
    assert attribute.name is "myint"
    assert attribute.is_primitive


def test_write_attribute_when_default():

    attribute = Attribute("myint", "integer", "description", optional=False, default=1)
    assert attribute.name is "myint"
    assert attribute.is_primitive
    blueprint = Blueprint(name="BP", package_path="")
    blueprint.attributes.append(attribute)
    entity = TestEntity(blueprint)
    entity.myint = 2
    writer = DMTWriter()
    res = writer.to_dict(entity)
    assert res.get("myint") is 2
    entity.myint = 1
    res = writer.to_dict(entity)
    assert (
        res.get("myint") is 1
    ), "The default value was used, but the attribute is not optional and should have been written"
    attribute.optional = True
    res = writer.to_dict(entity)
    assert "myint" not in res



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from pathlib import Path
import shutil

here = Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")


# Remove build and dist folders
shutil.rmtree(Path("build"), ignore_errors=True)
shutil.rmtree(Path("dist"), ignore_errors=True)

setup(
    name="dmtpy",
    version="0.1.0",
    author="SINTEF Ocean",
    description="Python runtime library for DMT",
    url="https://github.com/SINTEF/dmtpy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_dir={"": "src"},
    packages=find_packages(where="src", exclude=["tests"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)



"""
Blueprint with attributes
"""
from typing import Iterator, List
from .blueprint_attribute import BlueprintAttribute
from .dimension import Dimension
from .attribute import Attribute


class Blueprint:
    """The Entity metadata"""

    def __init__(self, name: str, package_path: str, description: str = None) -> None:
        self.name = name
        self.package_path = package_path
        self.version = 1
        self.description = description
        self.attributes: List[Attribute] = list()
        self.dimensions: List[Dimension] = list()

    def get_path(self):
        """Full path to this type"""
        return self.package_path + "/" + self.name

    def get_attribute(self, name: str) -> Attribute:
        """Get attribute of given name"""
        for prop in self.attributes:
            if prop.name == name:
                return prop
        return None

    def get_dimension(self, name: str) -> Dimension:
        """Get dimension by name"""
        for dim in self.dimensions:
            if dim.name == name:
                return dim
        return None

    def blueprint_attributes(self) -> Iterator[BlueprintAttribute]:
        """Get all blueprint attributes"""
        for p in self.attributes:
            if isinstance(p, BlueprintAttribute):
                yield p
